/*
 *
 * UAVCAN data structure definition.                                                                          +-+ +-+
 *                                                                                                            | | | |
 * AUTOGENERATED, DO NOT EDIT.                                                                                \  -  /
 *                                                                                                              ---
 *                                                                                                               o
 * +------------------------------------------------------------------------------------------------------------------+
 *
 * Generator:
 * nunavut-{{ nunavut.version }}
 *
 * Source File:
 * {{ T.source_file_path }}
 *
 * Template:
 * {{ self._TemplateReference__context.name }}
 *
 * Generated at:  {{ now_utc }} UTC
 * Is deprecated: {{ T.deprecated and 'yes' or 'no' }}
 * Fixed port ID: {{ T.fixed_port_id }}
 * Full name:     {{ T.full_name }}
 * Version:       {{ T.version.major }}.{{ T.version.minor }}
 *
 */

#ifndef {{T.full_name | c.macrofy}}
#define {{T.full_name | c.macrofy}}
{% for n in T | includes -%}
#include {{ n }}
{% endfor %}
{{T.full_namespace | open_namespace}}
{%- block object -%}{%- endblock -%}
{{T.full_namespace | close_namespace}}
#endif // {{T.full_name | c.macrofy}}

/*
{{ T | yamlfy }}
*/
